"""Message Interface."""

from typing import Any

from bitstring import BitStream

from renity.messages.interface import MessageMetaClass
from renity.serializers import Serializers as Serializer


class Message(metaclass=MessageMetaClass):
    """Message Interface.

    Args:
        _message(Any): Binary protocol message
    """

    __slots__ = ["_message", "data", "type", "serializer"]
    # prevent pytest from trying to discover tests in the class
    __test__ = False
    required: bool = False
    _length: int = 0

    def __init__(self, message: Any = None) -> None:
        self.message = message

    @property
    def message(self):
        """Message Dict."""
        return self._message

    @message.setter
    def message(self, value: Any) -> None:
        _message, data = self.__serialize(value)

        self._message = _message
        self.data = data

    def __serialize(self, message: Any) -> tuple:
        """Serialize Message.

        * Instantiate Serializer Chain on instance.

        Args:
            message(Any): message data.

        Returns:
            serialized_message(tuple): (<dict>, <bytes>)
        """
        # Instantiate Serializer Chain on instance.
        if not hasattr(self, "serializer"):
            _serializer = Serializer
            _serializer.message_cls = self
            self.serializer = _serializer()

        if message:
            self.serializer.message = message
            return self.serializer.run()

        return None, None

    def __iter__(self):
        """Generator Override.

        * Yields key/value pair from message dict.
        """
        yield from self.message.items()

    def __getitem__(self, __name: str) -> Any:
        """Dunder Override.

        * Enable Message object to be subscriptable.
        """
        return getattr(self, __name)

    def __bytes__(self):
        """Bytes Representation Override."""
        return BitStream(bytes=self.data).bytes



"""Validator Interface."""

from __future__ import annotations

import typing
from abc import ABCMeta
from abc import abstractmethod
from typing import Type


class ValidatorInterface(ABCMeta):
    """Validator Interface.

    The Validator interface declares a method for building the chain of validators.
    """

    @abstractmethod
    def _add(cls, validator: Validator) -> Validator:
        """Add Validator.

        Args:
            validator(Validator): next validator in chain.
        """
        pass

    @abstractmethod
    def verify(cls, *args: list) -> None:
        """Verify Field."""
        pass


class Validator(metaclass=ValidatorInterface):
    """Validator Base.

    Validation chaining is implemented within this base class.
    """

    _next: typing.Optional[Validator] = None

    def __init__(
        self, field: typing.Any, data_type: typing.Union[tuple, type]
    ) -> None:
        self._field = field
        self._data_type = data_type

    @classmethod
    def __subclasshook__(cls, subclass):  # noqa: D105
        return (
            hasattr(subclass, "add")
            and callable(subclass.add)
            and hasattr(subclass, "verify")
            and callable(subclass.verify)
            or NotImplemented
        )

    @property
    def add(self) -> typing.Callable:
        """Prevent accidental override."""
        return self.__add

    def __add(self, validator: Type[Validator]) -> None:
        """Add Validator.

        Args:
            validator(Validator): next validator in chain.
        """
        link = validator(field=self._field, data_type=self._data_type)
        pointer = self

        while pointer._next:
            pointer = pointer._next
        pointer._next = link

    def verify(self, request: typing.Any) -> bool:
        """Verify Field."""
        if self._next:
            return self._next.verify(request)
        return True



"""Renity Utils."""

from inspect import getmembers
from inspect import isclass
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type


def is_class_wrapper(func: Callable) -> Callable:
    """Return func if 'cls' is a class."""

    def wrapper(cls, *args):
        if not isclass(cls):
            return False
        return func(cls, *args)

    return wrapper


@is_class_wrapper
def subclassonly(cls: Type[Any], _type: type) -> bool:
    """Return whether 'cls' is derived from another class but is not the same class."""
    if issubclass(cls, _type) and not issubclass(_type, cls):
        return True
    return False


def modulesubclasses(__module: ModuleType, _type: Any) -> list:
    """Get Module Subclasses.

    Args:
        __module(ModuleType): Python Module
        _type(Any): Object Type

    Returns:
        list(key,value): name, and class
    """
    return getmembers(__module, lambda cls: subclassonly(cls, _type))


class Inventory:
    """Inventory Object.

    * Queue like dict object
    """

    def __init__(self):
        self.__value = {}

    def add(self, key: str, value: Any = None) -> None:
        """Add item to inventory.

        * Increment count
        * Append value to list
        """
        if key in self.__value:
            self.__value[key]["count"] += 1
            self.__value[key]["values"].append(value)
            return
        self.__value[key] = {"count": 1, "values": [value]}

    def pop(self, key: str) -> Optional[Any]:
        """Remove item to inventory.

        * Pop value from key->list
        * Remove key if list isEmpty
        """
        if key in self.__value:
            self.__value[key]["count"] -= 1
            item = self.__value[key]["values"].pop()
            if not self.__value[key]["count"]:
                self.__value.pop(key)
            return item
        return None

    def __len__(self) -> bool:
        """Not Empty?"""
        if self.__value:
            return True
        return False

    def __iter__(self):
        """Return all values."""
        for _, val in self.__value.items():
            yield from val["values"]



"""Field Validators.

Subclasses of type Validator
"""

from __future__ import annotations

from inspect import isclass

from renity.validators.interface import Validator

from ..utils import Inventory
from .exceptions import RequiredMessageField
from .exceptions import TooManyValues


class RequiredField(Validator):
    """Validate Required Field."""

    def verify(self, request):
        """Built-in Validation.

        * Verify required case
        """
        if request is None:
            raise RequiredMessageField(self._field.key, self._field.__class__)

        return super().verify(request)


class IncorrectFieldType(Validator):
    """Built-in Validation.

    * Verify correct type(value, default).
    """

    def verify(self, request):
        """Verify Field."""
        if request is not None and not isinstance(request, self._data_type):
            raise TypeError(
                f"Expected {self._data_type} but found {type(request)}."
            )

        return super().verify(request)


class SubFieldValidator(Validator):
    """Validate Packed List Elements.

    Ensure Packed LEN has is not empty.
    """

    def __init__(self, field, **_):
        self._field = field
        self._data_type = (list, tuple)

    def verify(self, request):
        """Validate Sub-Fields(list<Field>)."""
        # Verify Subfields
        if self._field.data_type is list:
            # Missing subfields
            if not len(self._field.sub_fields):
                raise TypeError(
                    "Missing at least 1 positional argument of type: <Field>"
                )

            # Expected subfield of type <Field>
            for field in self._field.sub_fields:
                sub_class = type(field)
                base = type(self._field).__mro__[1]

                if not isclass(sub_class) or not issubclass(sub_class, base):
                    raise TypeError(
                        "Invalid field expected type"
                        + f"{type(self._field).__mro__[1]}, but found type {field}"
                    )
        return super().verify(request)


class OverflowValidator(Validator):
    """Validate Packed List Elements.

    Ensure Packed LEN does not exceed size of sub_fields array.
    """

    def __init__(self, field, **_):
        self._field = field
        self._data_type = (list, tuple)

    def verify(self, request):
        """Verify Field."""
        subs = len(self._field.sub_fields)
        _input = len(request)
        if subs < _input:
            raise TooManyValues(subs, _input)
        return super().verify(request)


class UnorderedValidator(Validator):
    """Validate Unorderd Packed List Elements(experimental)."""

    def __init__(self, field, **_):
        self._field = field
        self._data_type = (list, tuple)

    def verify(self, request):
        """Verify Field."""
        valid = Inventory()
        for v in request:
            valid.add(str(type(v)), v)
        for sub in self._field.sub_fields:
            _type = str(sub.data_type)
            item = valid.pop(_type)
            sub.validate(item)

        if len(valid):
            raise TypeError(f"Unexpected value(s) of {list(valid)}")

        return super().verify(request)


class MessageTypeValidator(Validator):
    """Validate Message Type."""

    def verify(self, request):
        """Verify Field."""
        m_type = self._field.message_cls
        if not isinstance(request, str) or m_type != request:
            raise TypeError(
                f"Expected Message type: {m_type} but found {request}"
            )

        return super().verify(request)



"""Renity Constants."""

WIRE_TYPES = {
    # VARINT	int32, int64, uint32, uint64, sint32, sint64, bool, enum
    0: "_varint",
    # I64	fixed64, sfixed64, double
    1: "_i64",
    # LEN	string, bytes, embedded messages, packed repeated fields
    2: "_len",
    # MESSAGE IDENTIFIER -> LEN: String
    7: "_message_type",
}

# Decoder Constants

WIRE_MASK = 0b111

# Encoder Constants

FIELDS = {1: "001", 2: "010", 3: "011"}



"""Decoder Module Exceptions."""


class InvalidMessage(Exception):
    """Invalid Message Exception."""

    def __init__(self, _bits: str) -> None:
        _message = (
            "Message must begin with 8-Bit Message Identifier TLV \n"
            + f"expected 10010111 but found {_bits}"
        )  # noqa: B950
        self.message = _message
        self.code = 3101
        super().__init__(_message)



"""Renity Built-in Fields."""

from __future__ import annotations

from renity.validators.exceptions import IncorrectMessageType
from renity.validators.validators import MessageTypeValidator
from renity.validators.validators import OverflowValidator
from renity.validators.validators import SubFieldValidator

from .constants import BOOL
from .constants import FIXED64
from .constants import I64
from .constants import INT32
from .constants import LEN
from .constants import PACKED
from .constants import SINT32
from .constants import STR
from .constants import TYPE
from .constants import VARINT
from .interface import Field


__all__ = (
    "BoolField",
    "IntField",
    "StringField",
    "FloatField",
    "ListField",
)


class TypeField(Field):
    """Type Field.

    Default field that represents Message subclass name using LEN/string wire type.
    """

    wire = TYPE
    field = STR
    data_type = str

    def validate(self, value):
        """Explicit Built-in Field Validation."""
        validator = MessageTypeValidator(field=self, data_type=str)
        return validator.verify(value)


class IntField(Field):
    """Int Field.

    Built-in Field used to de/serialize Variable Integers of type/field(int32, sint64).

    Attributes:
        field(callable): returns type/field(int32, sint64).
    """

    wire = VARINT
    field = (INT32, SINT32)
    data_type = int

    def get_field_value(self):
        """Field Type From List.

        Returns:
            field(int): field type from list of options.

        Raises:
            IncorrectMessageType: Expected value of type <int>.
        """
        value = self.value
        try:
            if not value or value >= 0:
                field_type = 1
            else:
                field_type = 2
        except Exception as e:
            # Raise exception if not correct type
            raise IncorrectMessageType(
                type(value).__name__, int.__name__, value
            ) from e

        return field_type


class BoolField(Field):
    """Bool Field.

    Built-in Field used to de/serialize Variable Integers of type/field(bool).
    """

    wire = VARINT
    field = BOOL
    data_type = bool


class ListField(Field):
    """List Field.

    Built-in Field used to de/serialize LEN: Packed List.

    Attributes:
        sub_fields(list): list of subclass instances of type Field.
    """

    wire = LEN
    field = PACKED
    data_type = list
    validators = [
        SubFieldValidator,
        OverflowValidator,
    ]


class FloatField(Field):
    """Float Field.

    Built-in Field used to de/serialize I64: Fixed 64-bit float.
    """

    wire = I64
    field = FIXED64
    data_type = float


class StringField(Field):
    """String Field.

    Built-in Field used to de/serialize LEN: String.

    Attributes:
        sub_fields(list): list of subclass instances of type Field.
    """

    wire = LEN
    field = STR
    data_type = str



"""Field Interface.

Instances of this class are used to define a message's schema.
"""

from __future__ import annotations

from typing import Any
from typing import Union

from ..validators.exceptions import MissingPrimitiveException
from ..validators.interface import Validator
from ..validators.validators import IncorrectFieldType
from ..validators.validators import RequiredField
from ..validators.validators import UnorderedValidator


class FieldMeta(type):
    """Field metaclass.

    Used for field class creation.

    * Creates validation chain from 'validators' attribute
    """

    __field: Any = None
    field: Any = None

    def __new__(cls, name, bases, attrs):
        """Create new Field instance.

        * Initialize field property from field attribute
        """
        inst = super().__new__(cls, name, bases, attrs)
        if len(bases):
            inst.__field = attrs["field"]
            inst.field = cls._initialize_field()

            if not attrs["data_type"]:
                raise MissingPrimitiveException
        return inst

    @classmethod
    def _initialize_field(cls):
        def _field(self):
            if type(self.__field) in (list, tuple):
                return self.get_field_value()

            return self.__field

        def _set_field(self, value):
            self.__field = value

        return property(
            fget=_field,
            fset=_set_field,
            doc="""Wire Type Field Property.""",
        )

    def get_field_value(cls, *_):
        """Abstract Method.

        * Override for field type list or tuple
        """
        raise Exception(
            "Fields with multiple types must override this method."
        ) from NotImplementedError


class Field(metaclass=FieldMeta):
    """Binary Message Field Interface.

    Attributes:
        *wire(int): see 'constants.py' for valid wire types*

        *field(int, list<int>): **wire field or list of wire fields
        see 'constants.py' for valid wire fields.*

        **data_type(list, Any): Primitive or non-primitive data-type.**

        validators(Validator): see 'validators.py' for list of Validator(s).

    Args:


        key(str): name of field.

        value(Any): Set during serialization.

        required(bool): raise exception if field is missing in message.

        default(Any): default value of field.

        sub_fields(list): list of subclass instances (for packed wire type).

        sorted(bool): default=True flag for ordered list of
        subclasses **Warning: False(Experimental)**
    """

    data_type: Any = None
    validators: list = []
    field: Union[int, tuple[int, int]] = 0
    __value: Any = None
    key = ""
    message_cls: str = "Message"

    def __init__(
        cls,
        *sub_fields: Any,
        required: bool = False,
        default: Any = None,
        sorted: bool = True,
    ):
        cls.required = required
        cls.sorted = sorted
        cls.__initialize_sub_fields(sub_fields)
        cls._set_default_value(default)
        cls.__initialize_validator()

    @property
    def value(self):
        """Value Property."""
        return self.__value

    @value.setter
    def value(self, _val):
        sub_field_length = len(self.sub_fields)
        if sub_field_length:
            for idx in range(sub_field_length):
                self.sub_fields[idx].value = _val[idx]
        self.__value = _val

    @property
    def default(self):
        """Default Field Value."""
        return self.__default

    def _set_default_value(self, default: Any) -> None:
        # Validate default attribute
        if default and not isinstance(default, self.data_type):
            raise Exception(
                f"Expected default value type {self.data_type} but found {default}"
            ) from AttributeError

        self.__default = default

    def __initialize_sub_fields(self, sub_fields: Union[list, tuple]) -> None:
        if self.data_type is not list and len(sub_fields):
            name = self.__class__.__name__
            raise TypeError(
                f"{name} takse 0 positional arguments but {len(sub_fields)} was given."
            )
        self.sub_fields = sub_fields

    def __initialize_validator(self) -> None:
        data = self.data_type
        root = Validator(field=self, data_type=data)
        self.validator = root
        return self.__chain_validation()

    def __chain_validation(self) -> None:
        root: Validator = self.validator

        # RequiredField = built-in Validator
        if self.required:
            root.add(RequiredField)

        sub_fields = self.sub_fields

        if sub_fields:
            if not self.sorted:
                # Unsorted List Validation = built-in Validator *sub-fields*
                root.add(UnorderedValidator)
        elif not sub_fields or self.sorted:
            # Type Validation = built-in Validator/*sorted sub-fields validator*
            root.add(IncorrectFieldType)

        # Iterate Validator Subclasses
        if self.validators:
            for _validator_subclass in self.validators:
                if not issubclass(_validator_subclass, Validator):
                    raise TypeError(
                        f"TypeError: Expected type {Validator} but found {_validator_subclass}."
                    )
                root.add(_validator_subclass)

    def validate(self, value):
        """Field Validation.

        * Validate field
        * Traverse/validate sub_fields
        """
        results = [self.validator.verify(value)]
        if self.sorted:
            for idx, sub in enumerate(self.sub_fields):
                results.append(sub.validate(value[idx]))

        return all(results)



"""Renity Protocol Buffer.

Renity is an [Open Source] Python Object-Binary-Mapper(OBM)
Binary Protocol Buffer that provides a way to rapidly define
the de/serialization format of packets.

Advantages:
    · Improved throughput & reduced latency by reducing the data size       transferred over the network

    · Serialization & Deserialization

    · Efficiency Compared to other formats like JSON & XML

    · Strict schema definition(s) for messages

    · Backward Compatible Schema(s)

    · Development effort is reduced with an easy-to-use interface that creates a Serialization Format from an Object Model.


Example:
    >>> from renity import Message, StringField, IntField, ListField
    >>> class CustomMessage(Message):
    ...     hello=StringField(default="World")
    ...     sentence=ListField(StringField(required=True),IntField())
    ...

    >>> example = CustomMessage({"sentence": ["Number of Apples:", 2]})

    >>> example.message
    {'type': 'CustomMessage', 'hello': 'World', 'sentence': ['Number of Apples:', 2]}
"""

from .fields import fields
from .fields.fields import *  # noqa: F401, F403
from .messages.message import Message  # noqa: F401


__all__ = list(fields.__all__) + ["Message"]



"""Field Exceptions."""

from ..fields.constants import EMPTY_LIST_FIELD
from ..fields.constants import INCORRECT_MESSAGE_TYPE
from ..fields.constants import MISSING_PRIMITIVE
from ..fields.constants import REQUIRED_MESSAGE_FIELD
from ..fields.constants import TOO_MANY_VALUES


class MissingPrimitiveException(Exception):
    """Field Exception.

    Field Subclasses require primitive attribute.
    """

    def __init__(self):
        _message = "Field subclass missing primitive attribute declaration."
        super().__init__(_message)
        self.code = MISSING_PRIMITIVE
        self.message = _message


class RequiredMessageField(Exception):
    """Field Exception."""

    def __init__(self, name, required_field):
        _message = (
            f"Expected required message field {name}: {required_field.__name__}"
        )
        super().__init__(_message)
        self.code = REQUIRED_MESSAGE_FIELD
        self.message = _message


class IncorrectMessageType(Exception):
    """Field Exception."""

    def __init__(self, incorrect, correct, val):
        _message = f"Incorrect message type expected {correct} but found <{incorrect}> {val}"
        super().__init__(_message)
        self.code = INCORRECT_MESSAGE_TYPE
        self.message = _message


class InvalidField(Exception):
    """Field Exception."""

    def __init__(self):
        _message = "Invalid field expected type <Field>"
        super().__init__(_message)
        self.code = INCORRECT_MESSAGE_TYPE
        self.message = _message


class EmptyListField(Exception):
    """Field Exception."""

    def __init__(self):
        _message = "Implementation Exception: ListField must have sub-fields."
        super().__init__(_message)
        self.code = EMPTY_LIST_FIELD
        self.message = _message


class TooManyValues(Exception):
    """Field Exception."""

    def __init__(self, size: int, found: int):
        _message = f"TooManyValues: Expected {size} but found {found}."
        super().__init__(_message)
        self.code = TOO_MANY_VALUES
        self.message = _message



"""Message Serializer Interface."""

from __future__ import annotations

from abc import ABC
from typing import Any
from typing import Optional
from typing import Type


class MessageSerializer(ABC):
    """Message Serializer Interface.

    Attributes:
        message(dict): deserialized message.

        fields(list): <Message> sub-class fields.

        data(Any): Encoded message.

        data_type(type): data type of serializer.

        next(MessageSerializer): next node in Serializer Chain.
    """

    _message: dict = {}
    fields: list = []
    data: Optional[bytes] = None
    _next: Optional[MessageSerializer] = None
    length = 0
    message_length = 0

    @property
    def data_type(self) -> Type[Any]:
        """Abstract Data Type Property.

        * Force sub-classes implementation to include "data_type" attribute.
        """
        raise Exception(
            "Serializer sub-classes must have data_type attribute."
        ) from AttributeError

    @property
    def message(self) -> Optional[dict]:
        """Message getter."""
        return self._message

    @message.setter
    def message(self, _message: dict) -> None:
        """Message setter.

        Updates:
            - Message attributes
        """
        new_message = {}

        # Iterate Fields
        for idx in range(self.message_length):
            field = self.fields[idx]

            # <Field>, Key, Current Value, Bit
            f, key, __, bit = field

            # Get value from message

            # Check key
            _key_value = _message.get(key, None)

            # Check bit
            _bit_value = _message.get(bit, None)

            # Check Default
            _field_default = f.default

            value = next(
                (
                    item
                    for item in [
                        _key_value,
                        _bit_value,
                        _field_default,
                    ]
                    if item is not None
                ),
                None,
            )

            if value:
                f.validate(value)

            # Use key from field to update Dict
            new_message.update({key: value})

        self._message = new_message

    @property
    def next(self) -> Optional[MessageSerializer]:
        """Next Serializer Node."""
        return self._next

    @next.setter
    def next(self, _next_node: Any) -> None:
        self._next = _next_node

    def serialize(self, data: Any) -> None:
        """Serialize Method.

        * Must be overriden

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def load(self, data: Any) -> tuple:
        """Validate Message Processing.

        * Validate/Serialize Data or pass to next serializer for processing.

        Args:
            data(Any): Data to serialize

        Returns:
            (tuple): message, bytes.

        Raises:
            Exception: TypeError data-type serializer does not exist.
        """
        pointer: MessageSerializer = self

        if isinstance(data, self.data_type):
            self.serialize(data)
            return self.message, self.data

        if pointer.next:
            next: tuple = pointer.next.load(data)
            return next
        else:
            raise Exception(
                f"Serializer does not exist for type {type(data)}"
            ) from TypeError



"""Exception Codes.

Field Exception codes.
"""

INVALID_MESSAGE = 3101

REQUIRED_MESSAGE_FIELD = 3013

INCORRECT_MESSAGE_TYPE = 3014

EMPTY_LIST_FIELD = 3015

TOO_MANY_VALUES = 3016

MISSING_PRIMITIVE = 3017

"""Wire Types.

Wire Type Constants
"""
VARINT = 0  #: int32, int64, uint32, uint64, sint32, sint64, bool, enum
I64 = 1  #: fixed64, sfixed64, double
LEN = 2  #: string, bytes, embedded messages, packed repeated fields
TYPE = 7  #: MESSAGE IDENTIFIER -> LEN: String

"""Wire Fields.

Wire Field Constants
"""
BOOL = 3
PACKED = 1
INT32 = 1
STR = 2
FIXED64 = 1
SINT32 = 2



"""Message Interface Meta Module."""

from renity.fields.fields import TypeField
from renity.fields.interface import Field


class MessageMetaClass(type):
    """Message metaclass.

    Used for message class creation.

    * Creates validation chain from 'validators' attribute
    """

    def __new__(cls, name, bases, attrs):
        """Create new Message instance.

        * Initialize validators
        """
        _fields = {}
        _bits = {}
        _length = 0
        if not len(bases):
            cls._all_required = attrs["required"]

        def build(key, val, length):
            cls.strict(key, val)

            _fields.update({key: val})

            _bit = 2**length

            _bits.update({_bit: key})

        for _name, _value in attrs.items():
            if isinstance(_value, Field):
                build(_name, _value, _length)
                _length += 1

        # Add Type Field
        type_field = TypeField
        type_field.message_cls = name

        _fields.update({"type": type_field(default=name)})

        # Limit Message fields to 8bit
        if _length > 8:
            raise Exception(
                "Renity currently only supports 8 field Message Schema(s)."
            )

        attrs["_fields"] = _fields
        attrs["_bits"] = _bits
        attrs["_length"] = _length

        return super().__new__(cls, name, bases, attrs)

    @classmethod
    def strict(cls, key: str, field: Field) -> None:
        """Strict Message Schema.

        * Check Required True(flag all fields required=True)
        * Set key
        """
        if key == "type":
            raise TypeError("Attempted to overwrite protected field 'type'.")
        field.key = key
        if cls._all_required:
            field.required = True



"""Serializers."""

from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

from typeguard import typechecked

import renity
from renity.serializers import serializers
from renity.serializers.interface import MessageSerializer as Serializer

from ..utils import modulesubclasses


class FieldElement:
    """Serializer Field Object.

    Args:
        field(Field): Message Field.
        key(str): dict key.
        value(Any): Message property value.
        bit(int): starting bit of field.
    """

    __slots__ = ["field", "key", "_value", "bit", "_message"]

    @typechecked
    def __init__(
        self,
        field: renity.fields.Field,
        key: str,
        value: Any = None,
        bit: Optional[int] = None,
    ):
        self.field = field

        self.key = key

        self.value = value

        self.bit = bit

    @property
    def value(self):
        """Value Property."""
        return self._value

    @value.setter
    def value(self, val):
        # Set <Field> attribute value
        if val is not None:
            self.field.value = val

            # Validate Field Value
            self.field.validate(val)

        # Set value
        self._value = val if val is not None else self.field.default

    def __getitem__(self, __name: Any[str, int]) -> Any:
        """Dunder Override.

        * Enable Message object subscription.
        """
        return getattr(self, __name)

    def __iter__(self):
        """Dunder Override.

        * Iterable Attributes.
        """
        yield from [self.field, self.key, self.value, self.bit]


@typechecked
class Serializers:
    """Serializer Chain.

    Attributes:
        serializers(list): list of available <Serializer>(s).

        fields(list): list of <Message> sub-class field(s)<dict>

    Args:
        cls_name: Name of <Message> sub-class.

        message_cls: <Message> sub-class instance.
    """

    message_cls: Any
    _message: dict
    fields: list = []
    serializers: Serializer

    def __init__(self):
        """Chain Serializers."""
        self.chain()

    @property
    def message(self):
        """Message."""
        return self._message

    @message.setter
    def message(self, _message=None):
        self._message = _message
        if _message:
            self.load()
            self.update_chain()

    @property
    def next(self):
        """First link in Serializer Chain."""
        return self.serializers

    def load(self):
        """Load Message Subclass Attributes.

        Attributes:
            fields(list)<dict>:
                { bit(int), key(str), field(Field) }
        """
        m_cls = self.message_cls
        cls_fields = m_cls["_fields"]
        cls_bits = m_cls["_bits"]
        cls_length = m_cls["_length"]

        message = self.message if isinstance(self.message, dict) else {}

        # Get type field
        type_field = FieldElement(
            field=cls_fields["type"],
            key="type",
            value=message.get("type", None),
        )

        self.fields.append(type_field)

        for x in range(cls_length):
            # Pointer
            pointer = 2**x

            # Key
            key = cls_bits.get(pointer)

            # Field
            field = cls_fields[key]

            # dict
            element = FieldElement(
                field=field,
                key=key,
                bit=pointer,
                value=message.get(key, None),
            )

            self.fields.append(element)

    def chain(self) -> None:
        """Chain of Responsibility.

        Serializer Chain
            * Class definitions of type <Serializer> from serializers module.
        """
        for _, serializer in modulesubclasses(serializers, Serializer):
            self.add_link(serializer)

    def add_link(self, cls: Type[Serializer]) -> None:
        """Add Chain Link.

        * Set field(attr)
        * Set message max length: message_length(attr)
        * Instantiate Serializer class node.
        * Append to Linked List.

        Args:
            cls (Serializer): Next serializer in Linked List.
        """
        serializer: Type[Serializer] = cls
        serializer.fields = [list(field) for field in self.fields]

        # Defined Fields + built-in 'type' field
        serializer.message_length = self.message_cls._length + 1

        if not hasattr(self, "serializers"):
            self.serializers: Serializer = serializer()
        else:
            pointer = self.serializers

            while pointer.next:
                pointer = pointer.next
            pointer.next = serializer()

    def update_chain(self):
        """Update Serializer Chain.

        * Add Updated Fields to chain.
        """
        pointer: Serializer = self.serializers
        fields = [list(field) for field in self.fields]
        while pointer.next:
            pointer.fields = fields
            pointer: Optional[Callable] = pointer.next
        pointer.fields = fields

    def run(self) -> tuple:
        """Run Method.

        * Traverse serializers process or pass to next.

        Returns:
            serialized_data(tuple): (<dict>, <bytes>)
        """
        _message, _data = self.serializers.load(self.message)

        return self.serializers.load(self.message)



"""Decoder Module."""

import codecs
import typing

from bitstring import Bits
from bitstring import BitStream

from ..constants import WIRE_MASK
from ..constants import WIRE_TYPES
from .exceptions import InvalidMessage


# Static Module Message Decoder


# Module Vars
bits: BitStream = BitStream()
decoded_bytes: dict = dict()
length: int = 0
attributes: list = []
decoded_value: dict = dict()
# Module Vars


def decode(_bits):
    """Decode Message."""
    global bits
    global decoded_bytes
    global length
    global attributes
    global decoded_value

    # Message Start
    # Assert Message begins with 'Message Type(int): 7'
    bits = BitStream(bytes=_bits)
    decoded_bytes = bits.bytes
    length = len(bits) - 1
    attributes = []
    decoded_value = {}

    # Get Message Type For Constructing Message
    message_type = bits.read(8)

    # Raise Exception if Message does not begin with Identifier
    if message_type.int & WIRE_MASK != 7:
        raise InvalidMessage(message_type.bin)

    # Get Message Type from LEN: String Protocol
    message_type, _ = base_len(data=bits, field=2)

    # Update Decoded Dict with message_type
    decoded_value["type"] = message_type

    # Read Attributes included in message
    for idx, x in enumerate(bits.read(8).bin):
        # Attributes Included/Absent On/Off = 1/0
        if x == "1":
            attributes = [2 ** abs(idx - 7)] + attributes

    # Empty Message will not throw exception, but return message containing only type
    if bits.pos == len(bits):  # pragma: no cover
        return decoded_value

    # Wire Type of first attribute to decode
    start_wire_protocol = WIRE_TYPES[bits.peek(8).int & WIRE_MASK]

    # set next wire protocol
    next_wire = globals()[start_wire_protocol]

    # While message has next wire protocol decode
    while next_wire:
        # Decode next message
        value, _ = next_wire()

        # Get message attribute-key(int)
        key = next_attr()

        # Add decoded message to dict
        decoded_value[key] = value

        # Get next wire protocol
        next_wire = advance()

    return decoded_value


def bytes(_bits):
    """Byte representation of decoded data."""
    global decoded_bytes
    decode(_bits)
    return decoded_bytes


def tag() -> tuple:
    """Message Attribute TLV.

    Returns:
        TLV(tuple): (Wire Type(str), Wire Field(int))
    """
    global bits

    # Peek at TLV
    _tag = bits.peek(8)

    # Get Wire Type from TLV
    _wire = WIRE_TYPES[_tag.int & WIRE_MASK]

    # Get Wire Field from TLV
    _field = _tag[1:5].int

    # Wire is Message Identifier ignore tag(set 2)
    if _wire == "_message_type":  # pragma: no cover
        _field = 2

    return (_wire, _field)


def advance():
    """Advance pointer.

    * To next message in Binary Protocol

    Returns:
        Next protocol or None

    Raises:
        Exception: General Exception from advancing bit pointer.
    """
    global bits
    global length

    try:
        # Check if there are bits left
        if bits.pos < length:
            # Get tag from current position
            wire, field = tag()

            # Return appropriate protocol for wire_type
            return globals()[wire]

        # End of bytes
        return None
    except Exception as e:  # pragma: no cover
        raise e


def next_attr():
    """Pop next attribute from queue."""
    global attributes
    return attributes.pop(0)


def base_wrapper(base_func: typing.Callable) -> typing.Callable:
    """Deserialization Base Method Wrapper.

    Args:
        base_func(callable): Base Wire Deserialization Method.

    Returns:
        (tuple): Deserialized Value, Next Wire Method Call
    """

    def wrapper(*args, **kwargs):
        global bits
        # Get Wire Field Type
        _, field = tag()

        # Advance Pointer past TLV
        _bin_tag = bits.read(8)  # noqa: F841

        # Base Method Call
        value, _length = base_func(*args, data=bits, field=field, **kwargs)

        # Return Value & Next Wire Method Call
        return value, advance()  # can add _length

    return wrapper


def __get_stream(data: typing.Union[BitStream, str]) -> BitStream:
    if isinstance(data, BitStream):
        _bits = data
    else:
        _bits = BitStream(bin=data)
    return _bits


def base_varint(data, field=0, *args, **kwargs):
    """VarInt.

    * Variable length integer

    Args:
        data(BitStream, bytes, bin...): Binary Data
        field(int): Field type of VARINT wire type.
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Returns:
        value(int): Variable Integer.
        length(int): Length of Wire.
    """
    _bits = __get_stream(data)

    def zig_zag(i):  # pragma: no cover
        # Signed int ZigZag Decoder (field[1] == sint32)
        return (i >> 1) ^ -(i & 1)

    # Initialize Bit stream from bits arg

    value: typing.Union[Bits, BitStream, int] = BitStream()

    # Initialize Most important bit
    msb = 1

    # Initialize length of varint
    length = 0

    # While MSB == 1 continue
    while msb:  # pragma: no cover
        # read next 8-bit sequence
        bin = _bits.read(8)

        # set MSB to current 8-bit sequence
        msb = bin.int >> 7

        # Get remaining 7-bits
        bin = bin[1:]

        # Convert sequence to big-endian
        value = bin + value if value.any(0) or value.any(1) else bin

        # Add 8-bits to length
        length += 8

        # MSB is 0 == end of varint
        if not msb:
            break

    value = int(f"0b{value.bin}", 2)

    # Value is sint32
    if field == 2:
        value = zig_zag(value)
    # Value is bool
    elif field == 3:
        value = bool(value)

    return (value, length)


@base_wrapper
def _varint(*args, **kwargs):
    """Var Int Advance Method.

    * Advance Message Deserialization and return value.

    Args:
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Returns:
        Value/Function call into next Wire Type for deserialization.
    """
    return base_varint(*args, **kwargs)


def base_i64(data, *args, **kwargs):
    """I64 (Float).

    * Base call for unpacking 64-bit Floats

    Args:
        data(BitStream, bytes, bin...): Binary Data
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Returns:
        (tuple): value,length
    """
    _bits = __get_stream(data)

    float = _bits.read("float64")
    return (float, 64)


@base_wrapper
def _i64(*args, **kwargs):
    """I64 Advance Method.

    * Advance Message Deserialization and return value.
    * 8 bytes for IEEE 754 binary64

    Args:
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Returns:
        Value/Function call into next Wire Type for deserialization.
    """
    return base_i64(*args, **kwargs)


def base_len(data, field, *args, **kwargs):
    """Length Delimited Record.

        - 8bit(Tag)
        - 8bit(Int32) Length = Length * 8

    Args:
        data(BitStream, bytes, bin...): Binary Data
        field(int): Field type of VARINT wire type.
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Raises:
        AttributeError: Invalid Field type.

    Returns:
        Case 1:
            List of Primitive Scalar Types
        Case 2:
            Decoded utf-8 String
    """
    _bits = __get_stream(data)

    # Advance pointer past LEN's length/Int32 TLV(8bits)
    _bits.read(8)  # noqa: F841

    # Decode Length Delimited Records Length(Int32)
    length, _int_len = base_varint(_bits[_bits.pos :].bin)

    # Advance pointer past Length
    _bits.read(_int_len)

    if field == 1:
        # LEN: Packed - Unpack List
        value: typing.Any = unpack(length * 8)
    elif field == 2:
        # LEN: String decode utf-8 String
        value = _bits.read(length * 8)

        value = codecs.decode(value.bytes, "utf-8")

    else:
        raise AttributeError("LEN: Field does not exist.")

    return value, (length * 8) + _int_len + 8


@base_wrapper
def _len(*args, **kwargs):
    """LEN Advance Method.

    * Advance Message Deserialization and return value.
    * Length Delimited Record

    Args:
        *args(Any): Optional*
        **kwargs(Any): Optional**

    Returns:
        Value/Function call into next Wire Type for deserialization.
    """
    return base_len(*args, **kwargs)


def unpack(_length: int) -> list:
    """LEN: PACKED.

    * Packed list of Primitives.
    * Not list, tuple, dict, bytes.

    Args:
        _length(int): Length of bits that represent list in message

    Returns:
        unpacked(list): List of Primitive Scalar Types

    Raises:
        TypeError: Expects, cannot nest list, tuple, dict etc.
    """
    global bits

    unpacked = []

    # End of Message
    end = bits.pos + _length

    # Continue decoding until end of message
    while end > bits.pos:
        # Get wire_type of value
        wire, field = tag()

        if wire == 2 and field == 1:
            raise TypeError(
                "TypeError: Expected primitive, cannot nest data structures in packed list."
            )

        # Advance pointer past TLV of value
        bits.read(8)

        # Get value of primitive from wire function
        value, pointer = globals()[f"base{wire}"](bits[bits.pos :].bin, field)

        # Advance pointer to start of next value
        bits.read(pointer)

        # Append decoded value to list
        unpacked.append(value)

    return unpacked



"""Message Serializer Sub-Classes."""

from typing import Optional

from renity.decoder import decoder as MessageDecoder
from renity.encoder.encoder import Encoder
from renity.serializers.interface import MessageSerializer


class DictionarySerializer(MessageSerializer):
    """Dictionary Serializer.

    * for <Message> of type <dict>
    """

    data_type = dict

    def serialize(self, message: Optional[dict] = None) -> None:
        """Method Override."""
        self.message = message
        self.data = Encoder.encode(self.fields)


class ByteSerializer(MessageSerializer):
    """Byte Serializer.

    * for <Message> of type <str>(bits)
    """

    data_type = bytes

    def serialize(self, data: bytes) -> None:
        """Method Override."""
        self.data = data
        # Decode message bytes -> dict
        self.message = MessageDecoder.decode(data)



"""Encoder Module."""

import typing
from math import ceil
from typing import Any

from bitstring import Bits
from bitstring import BitStream

from renity.fields.fields import ListField

from ..constants import FIELDS
from ..constants import WIRE_TYPES


class Encoder:
    """Encoder.

    Attributes:
        _varint(dict): valid types (int32,sint32,bool)
        _i64(dict): valid types (fixed64)
        _len(dict): valid types (list(packed), string)
        _message_type(dict): string
    """

    _varint = {1: "_int32", 2: "_sint32", 3: "_bool"}
    _i64 = {1: "fixed64"}
    _len = {1: "_packed", 2: "_string"}
    _message_type = {2: "_string"}

    @classmethod
    def encode(cls, _fields: list) -> bytes:
        """Encode Message.

        Args:
            _fields(list): field(Field), key(str), value(Any), bit(int) Quartet

        Returns:
            _bits(str): binary representation of encoded values
                - Attributes(8bit) + Records(>0bits)
        """
        # Attibutes 8-bit representation of 8 fields
        attributes = 0
        # Intialize encoded records list
        records = []
        # Initialize Message Identifier
        identifier = ""

        # iterate fields + encode
        for field, _, value, bit in _fields:
            if value is not None:
                # Get wire type from Field
                _wire_type = WIRE_TYPES[field.wire]

                # Get wire field from Field
                _wire_field = field.field

                # Get corresponding encoder protocol name
                wire = getattr(cls, _wire_type)[_wire_field]

                # Get corresponding encoder protocol from name
                _encoder = getattr(cls, wire)

                # Encode field
                record = _encoder(value, field, f"{field.wire:0b}".zfill(3))

                # message_type(Used for constructing messages does not have a bit)
                # set and continue loop
                if bit is None:
                    identifier = record
                    continue

                # Append encoded field to records list
                records.append(record)

                # Mark corresponding attribute bit
                attributes += bit

        result: bytes = BitStream(
            bin=identifier + f"{attributes:0b}".zfill(8) + "".join(records)
        ).bytes
        return result

    def __getitem__(self, wire: str) -> Any:
        """Override.

        - Makes class instance subscriptable.

        """
        return self.__getattribute__(wire)

    @classmethod
    def _bool(
        cls, _value: bool, _: typing.Any = None, __: typing.Any = None
    ) -> str:
        """Boolean Protocol.

        Args:
            _value(bool)

        Returns:
            (str): Bool bit-str
        """
        # Generate Tag
        tag = "1" + f"0{FIELDS[3]}000"
        # Convert bool to 8bit string
        value = str(int(_value)).zfill(8)

        return tag + value

    @classmethod
    def varint(
        cls,
        value: int,
        *args: typing.Any,
    ) -> str:
        """Encode Variable Int.

        Args:
            value(int): integer to be encoded
            args(list): arbitrary args

        Returns:
            (str): Binary String representation of integer.
        """
        _bin = f"{value:0b}"
        _len = len(_bin)
        _bytes = ceil(_len / 7)
        _b = []
        for x in range(_bytes):
            right = _len - x * 7
            left = right - 7 if right - 7 > 0 else 0
            _b.append(
                f"{1 if left != 0 else 0}{_bin[left:right if right > 0 else 1].zfill(7)}"
            )
        return "".join(_b)

    @classmethod
    def _sint32(
        cls,
        _value: int,
        *args: typing.Any,
    ) -> str:
        """Encode Signed Variable Int.

        Args:
            _value(int): integer to be encoded
            args(list): arbitrary args

        Returns:
            (str): Binary String representation of signed integer.
        """
        # ZigZag encode signed int
        value = (_value << 1) ^ (_value >> 31)
        # Generate Binary Protocol Message from TLV + Varint
        _bits = "1" + f"0{FIELDS[2]}000" + cls.varint(value)
        return _bits

    @classmethod
    def _int32(
        cls,
        value: int,
        *args: typing.Any,
    ) -> str:
        """Encode Variable Int.

        Args:
            value(int): integer to be encoded
            args(list): arbitrary args

        Returns:
            (str): Binary String representation of integer.
        """
        # Generate Binary Protocol Message from TLV + Varint
        _bits = "1" + f"0{FIELDS[1]}000" + cls.varint(value)
        return _bits

    @classmethod
    def fixed64(
        cls,
        value: float,
        *args: typing.Any,
    ) -> str:
        """64bit Float."""
        # Encode 64-bit Float value
        _bits = Bits(float=value, length=64)

        # Return Binary Protocol Message From TLV + 64-bit Float
        result: str = "1" + f"{FIELDS[1]}001".zfill(7) + _bits.bin
        return result

    @classmethod
    def _string(cls, value: str, _: Any = None, wire: str = "010") -> str:
        """LEN: String.

        Args:
            value(str): value to encode.
            wire(str): wire key.

        Returns:
            (str): result
        """
        # Generate Tag
        _tag = "1" + f"0{FIELDS[2]}{wire}"

        # Convert string to bits
        _bits = "".join(format(ord(i), "08b") for i in value)

        # Get 8bit length
        _length = cls._int32(int(len(_bits) / 8))

        # Return Binary Protocol for LEN: String
        return _tag + _length + _bits

    @classmethod
    def _packed(cls, values: list, field: ListField, __: Any = None) -> str:
        """Packed value(list).

        Args:
            values(list): list of (any scalar type that is not sting/bytes)
            field(ListField): iterate sub_fields in ListField

        Returns:
            (list): result

        Raises:
            Exception: General Exception
        """
        # Generate TLV
        _tag = "1" + f"0{FIELDS[1]}010"

        # Initialize Bits String
        _bits = ""
        try:
            # Iterate Subfields
            for idx, sub in enumerate(field.sub_fields):
                try:
                    # Get TLV Field
                    _wire_field = sub.field
                except IndexError:
                    # If value is not in list continue to next iteration
                    continue

                # Get wire_type
                _wire_type = WIRE_TYPES[sub.wire]

                # Get encoder func name [wire_type]: {[tlv_field]: {...}}
                wire = getattr(cls, _wire_type)[_wire_field]

                # Get encoder func
                _encoder = getattr(cls, wire)

                # Call encoder func
                record = _encoder(values[idx], field)

                # Append returned record
                _bits += record

            # Return Binary Protocol for LEN: Packed List
            return _tag + cls._int32(int(len(_bits) / 8)) + _bits
        except Exception as e:
            raise e
